/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          scanner.h
**   Purpose:       Scanner header file
**   Programmer:    Kevin Rosenberg
**   Date Started:  July 1, 1984
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: scanner.h,v 1.3 2000/06/19 17:58:20 kevin Exp $
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#ifndef SCANNER_H
#define SCANNER_H

#include "projections.h"

// Projections are collected along an array of ndet detectors.  The data
// for these detectors is stored in the class DetectorArray

typedef float DetectorValue;

class DetectorArray
{
 public:
  DetectorArray (const int ndet);
  ~DetectorArray (void);

  const int nDet(void) const {return m_nDet;}
  const double viewAngle(void) const {return m_viewAngle;}
  DetectorValue* detValues(void) {return m_detValues;}
  const DetectorValue* detValues(void) const {return m_detValues;}

  void setViewAngle (double viewAngle)
      { m_viewAngle = viewAngle; }

 private:
  DetectorValue* m_detValues;	/* Pointer to array of values recorded by detector */
  int m_nDet;		/* Number of detectors in array */
  double m_viewAngle;    /* View angle in radians */

  DetectorArray& operator=(const DetectorArray& rhs);
  DetectorArray& operator()(const DetectorArray& rhs);
};

typedef enum {
    DETECTOR_PARALLEL,
    DETECTOR_EQUIANGLE,
    DETECTOR_EQUILINEAR
} ScannerGeometry;
  

class Scanner
{
 public:
  Scanner (const Phantom& phm, const ScannerGeometry geometry, int nDet, int nView, int nSample, const double rot_anglen);
  ~Scanner();
  
  void collectProjections (Projections& proj, const Phantom& phm, const int start_view, const int trace);

  void setNView (int nView);
  const unsigned int nDet(void) const {return m_nDet;}
  const unsigned int nView(void) const {return m_nView;}
  const double phmLen(void) const {return m_phmLen;}
  const double rotInc(void) const {return m_rotInc;}
  const double detInc(void) const {return m_detInc;}
  const double radius(void) const {return m_radius;}


 private:
  void projectSingleView (const Phantom& phm, DetectorArray& darray, const double xd1, const double yd1, const double xd2, const double yd2, const double xs1, const double ys1, const double xs2, const double ys2);

  double projectSingleLine (const Phantom& phm, const double x1, const double y1, const double x2, const double y2);

  double projectLineAgainstPElem (const PhantomElement& pelem, const double x1, const double y1, const double x2, const double y2);

  void traceShowParam (const char *label, const char *fmt, int row, int color, ...);


  ScannerGeometry m_geometry;     /* Geometry of detectory */
  unsigned int m_nDet;			/* Number of detectors in array */
  unsigned int m_nView;			/* Number of rotated views */
  unsigned int m_nSample;		        /* Number of rays per detector */
  double m_detLen;		/* Total length of detector array */
  double m_rotLen;		/* Rotation angle length in radians (norm 2PI) */
  double m_detInc;		/* Increment between centers of detectors */
  double m_rotInc;		/* Increment in rotation angle between views */
  double m_radius;		/* Radius of rotation.  Distance from */
				/*   center of phm to center of det */
  double m_phmLen;                /* Maximum Length of phantom or area of interest */

  struct {
    double xd1,yd1,xd2,yd2;	/* Coordinates of detector endpoints */
    double xs1,ys1,xs2,ys2;	/* Coordinates of source endpoints */
    double angle;		/* Starting angle */
  } m_initPos;

  static const int N_EXTRA_DETECTORS=4;           /* Number of extra detectors widths when calculating detlen */

  int m_trace;

};

const static int RAYSUM_TRACE_ROW_TITLE=1;
const static int RAYSUM_TRACE_ROW_TITLE2=2;
const static int RAYSUM_TRACE_ROW_PHANT_ID=4;
const static int RAYSUM_TRACE_ROW_CHROMATIC=7;
const static int RAYSUM_TRACE_ROW_SCATTER=8;
const static int RAYSUM_TRACE_ROW_PHOT_STAT=9;
const static int RAYSUM_TRACE_ROW_NDET=12;
const static int RAYSUM_TRACE_ROW_NVIEW=13;
const static int RAYSUM_TRACE_ROW_SAMPLES=14;
const static int RAYSUM_TRACE_ROW_CURR_VIEW=17;
const static int RAYSUM_TRACE_ROW_ATTEN=18;



#endif
