/*****************************************************************************
** FILE IDENTIFICATION
** 
**     Name:                   filter.cpp
**     Purpose:                Routines for signal-procesing filters
**     Progammer: 	       Kevin Rosenberg
**     Date Started:           Aug 1984
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: filter.cpp,v 1.1 2000/06/19 02:59:34 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"


/* NAME
 *   filter_generate				Generate a filter
 *
 * SYNOPSIS
 *   f = filter_generate (filt_type, bw, xmin, xmax, n, param, domain, analytic)
 *   double f				Generated filter vector
 *   int filt_type			Type of filter wanted
 *   double bw				Bandwidth of filter
 *   double xmin, xmax			Filter limits
 *   int n				Number of points in filter
 *   double param			General input parameter to filters
 *   int domain				FREQ or SPATIAL domain wanted
 *   int numint				Number if intervals for calculating
 *					discrete inverse fourier xform
 *					for spatial domain filters.  For
 *					ANALYTIC solutions, use numint = 0
 */

double *
filter_generate (const FilterType filt_type, double bw, double xmin, double xmax, int n, double param, const DomainType domain, int numint)
{
  double *f = new double [n];
  double xinc = (xmax - xmin) / (n - 1);

  if (filt_type == FILTER_SHEPP) {
    double a = 2 * bw;
    double c = - 4. / (a * a);
    int center = (n - 1) / 2;
    int sidelen = center;
    f[center] = 4. / (a * a);
    
    for (int i = 1; i <= sidelen; i++ )
      f [center + i] = f [center - i] = c / (4 * (i * i) - 1);
  } else if (domain == D_FREQ) {
    double x;
    int i;
    for (x = xmin, i = 0; i < n; x += xinc, i++)
      f[i] = filter_frequency_response (filt_type, x, bw, param);
  } else if (domain == D_SPATIAL) {
    double x;
    int i;
    for (x = xmin, i = 0; i < n; x += xinc, i++)
      if (numint == 0)
	f[i] = filter_spatial_response_analytic (filt_type, x, bw, param);
      else
	f[i] = filter_spatial_response_calc (filt_type, x, bw, param, numint);
  } else {
    sys_error (ERR_WARNING, "Illegal domain %d [filt_generate]", domain);
    return (NULL);
  }
  
  return (f);
}


/* NAME
 *   filter_spatial_response_calc	Calculate filter by discrete inverse fourier
 *				        transform of filters's frequency
 *  		 		        response
 *
 * SYNOPSIS
 *   y = filter_spatial_response_calc (filt_type, x, bw, param, n)
 *   double y			Filter's response in spatial domain
 *   int filt_type		Type of filter (definitions in ct.h)
 *   double x			Spatial position to evaluate filter
 *   double bw			Bandwidth of window
 *   double param		General parameter for various filters
 *   int n			Number of points to calculate integrations
 */

double 
filter_spatial_response_calc (int filt_type, double x, double bw, double param, int n)
{
  double zmin, zmax;

  if (filt_type == FILTER_TRIANGLE) {
    zmin = 0;
    zmax = bw;
  } else {
    zmin = 0;
    zmax = bw / 2;
  }
  double zinc = (zmax - zmin) / (n - 1);

  double z = zmin;
  double q [n];
  for (int i = 0; i < n; i++, z += zinc)
    q[i] = filter_frequency_response (filt_type, z, bw, param) * cos (TWOPI * z * x);
  
  double y = 2 * integrateSimpson (zmin, zmax, q, n);
  
  return (y);
}


/* NAME
 *    filter_frequency_response			Return filter frequency response
 *
 * SYNOPSIS
 *    h = filter_frequency_response (filt_type, u, bw, param)
 *    double h			Filters frequency response at u
 *    int filt_type		Type of filter
 *    double u			Frequency to evaluate filter at
 *    double bw			Bandwidth of filter
 *    double param		General input parameter for various filters
 */

double 
filter_frequency_response (int filt_type, double u, double bw, double param)
{
  double q;
  double au = fabs (u);

  switch (filt_type) {
  case FILTER_BANDLIMIT:
    if (au >= bw / 2)
      q = 0.;
    else
      q = 1;
    break;
  case FILTER_ABS_BANDLIMIT:
    if (au >= bw / 2)
      q = 0.;
    else
      q = au;
    break;
  case FILTER_TRIANGLE:
    if (au >= bw)
      q = 0;
    else
      q = 1 - au / bw;
    break;
  case FILTER_COSINE:
    if (au >= bw / 2)
      q = 0;
    else
      q = cos(PI * u / bw);
    break;
  case FILTER_ABS_COSINE:
    if (au >= bw / 2)
      q = 0;
    else
      q = au * cos(PI * u / bw);
    break;
  case FILTER_SINC:
    q = bw * sinc (PI * bw * u, 1.);
    break;
  case FILTER_ABS_SINC:
    q = au * bw * sinc (PI * bw * u, 1.);
    break;
  case FILTER_G_HAMMING:
    if (au >= bw / 2)
      q = 0;
    else
      q = param + (1 - param) * cos (TWOPI * u / bw);
    break;
  case FILTER_ABS_G_HAMMING:
    if (au >= bw / 2)
      q = 0;
    else
      q = au * (param + (1 - param) * cos(TWOPI * u / bw));
    break;
  default:
    q = 0;
    sys_error (ERR_WARNING,
	       "Frequency response for filter %d not implemented [filter_frequency_response]",
	       filt_type);
    break;
  }
  return (q);
}



/* NAME
 *   filter_spatial_response_analytic			Calculate filter by analytic inverse fourier
 *				transform of filters's frequency
 *				response
 *
 * SYNOPSIS
 *   y = filter_spatial_response_analytic (filt_type, x, bw, param)
 *   double y			Filter's response in spatial domain
 *   int filt_type		Type of filter (definitions in ct.h)
 *   double x			Spatial position to evaluate filter
 *   double bw			Bandwidth of window
 *   double param		General parameter for various filters
 */

double 
filter_spatial_response_analytic (int filt_type, double x, double bw, double param)
{
  double q, temp;
  double u = TWOPI * x;
  double w = bw / 2;
  double b = PI / bw;
  double b2 = TWOPI / bw;

  switch (filt_type) {
  case FILTER_BANDLIMIT:
    q = bw * sinc(u * w, 1.0);
    break;
  case FILTER_TRIANGLE:
    temp = sinc (u * w, 1.0);
    q = bw * temp * temp;
    break;
  case FILTER_COSINE:
    q = sinc(b-u,w) + sinc(b+u,w);
    break;
  case FILTER_G_HAMMING:
    q = 2 * param * sin(u*w)/u + (1-param) *
      (sinc(b2-u, w) + sinc(b2+u, w));
    break;
  case FILTER_ABS_BANDLIMIT:
    q = 2 * integral_abscos (u, w);
    break;
  case FILTER_ABS_COSINE:
    q = integral_abscos(b-u,w) + integral_abscos(b+u,w);
    break;
  case FILTER_ABS_G_HAMMING:
    q = 2 * param * integral_abscos(u,w) +
      (1-param)*(integral_abscos(u-b2,w)+integral_abscos(u+b2,w));
    break;
  case FILTER_SHEPP:
    if (fabs (u) < 1E-7)
      q = 4. / (PI * bw * bw);
    else
      q = fabs ((2 / bw) * sin (u * w)) * sinc (u * w, 1.) * sinc (u * w, 1.);
    break;
  case FILTER_SINC:
    if (fabs (x) < bw / 2)
      q = 1.;
    else
      q = 0.;
    break;
  case FILTER_ABS_SINC:
  default:
    sys_error (ERR_WARNING,
	       "Analytic filter type %d not implemented [filter_spatial_response_analytic]",
	       filt_type);
    q = 0;
    break;
  }
  
  return (q);
}


/* NAME
 *   sinc			Return sin(x)/x function
 *
 * SYNOPSIS
 *   v = sinc (x, mult)
 *   double v			sinc value
 *   double x, mult
 *
 * DESCRIPTION
 *   v = sin(x * mult) / x;
 */

double 
sinc (double x, double mult)
{
  return (fabs(x) > F_EPSILON ? (sin (x * mult) / x) : 1.0);
}


/* NAME
 *   integral_abscos			Returns integral of u*cos(u)
 *
 * SYNOPSIS
 *   q = integral_abscos (u, w)
 *   double q			Integral value
 *   double u			Integration variable
 *   double w			Upper integration boundary
 *
 * DESCRIPTION
 *   Returns the value of integral of u*cos(u)*dV for V = 0 to w
 */

double 
integral_abscos (double u, double w)
{
  if (fabs (u) > F_EPSILON)
    return (cos(u * w) - 1) / (u * u) + w / u * sin (u * w);
  else
    return (w * w / 2);
}


