/*****************************************************************************
** FILE IDENTIFICATION
** 
**     Name:                   phm.cpp
**     Purpose:                Routines for phantom objects
**     Progammer: 	       Kevin Rosenberg
**     Date Started:           Aug 1984
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: phantom.cpp,v 1.1 2000/06/19 02:59:34 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"


// CLASS IDENTIFICATION
//   Phanton
//

Phantom::Phantom (void)
{
  m_nPElem = 0;
  m_xmin = 1E30;
  m_xmax = -1E30;
  m_ymin = 1E30;
  m_ymax = -1E30;
  m_diameter = 0;
  m_composition = P_PELEMS;
}


Phantom::~Phantom (void)
{
  for (PElemIterator i = m_listPElem.begin(); i != m_listPElem.end(); i++) {
    delete *i;
  }
}


void
Phantom::create (const int phmid)
{
  switch (phmid) 
    {
    case O_PHM_HERMAN:
      std_herman();
      break;
    case O_PHM_ROWLAND:
      std_rowland();
      break;
    case O_PHM_BROWLAND:
      std_rowland_bordered ();
      break;
    case O_PHM_UNITPULSE:
      m_composition = P_UNIT_PULSE;
      addPElem ("rectangle", 0., 0., 100., 100., 0., 0.);     // outline 
      addPElem ("ellipse", 0., 0., 1., 1., 0., 1.);	      // pulse 
      break;
    default:
      sys_error (ERR_WARNING, "Illegal phantom id %d\n", phmid);
      break;
    }
}


/* METHOD IDENTIFICATION
 *   createFromFile          Add PhantomElements from file
 *
 * SYNOPSIS
 *   createFromFile (filename)
 *
 * RETURNS
 *   true if pelem were added
 *   false if an pelem not added
 */

bool
Phantom::createFromFile (const char* const fname)
{
  bool stoploop = false;
  bool retval = false;
  FILE *fp;

  if ((fp = fopen (fname, "r")) == NULL)
    return (false);

  do {
    double cx, cy, u, v, rot, dens;
    char pelemtype[80];
    int n = fscanf (fp, "%79s %lf %lf %lf %lf %lf %lf",
		pelemtype, &cx, &cy, &u, &v, &rot, &dens);
    
    if (n == EOF || n == 0) {	/* end of file */
      stoploop = true;
      retval = false;
    } else if (n != 7) {
      stoploop = true;
      retval = false;
    } else {
      addPElem (pelemtype, cx, cy, u, v, rot, dens);
      retval = true;
    }
  } while (stoploop == false);
  
  fclose (fp);

  return (retval);
}


/* NAME
 *   addPElem		Add pelem
 *
 * SYNOPSIS
 *   addPElem (type, cx, cy, u, v, rot, atten)
 *   char *type		type of pelem (box, ellipse, etc)
 *   double cx, cy	pelem center
 *   double u,v		pelem size
 *   double rot		rotation angle of pelem (in degrees)
 *   double atten	x-ray attenuation cooefficient
 */

void 
Phantom::addPElem (const char *type, const double cx, const double cy, const double u, const double v, const double rot, const double atten)
{
  PhantomElement *pelem = new PhantomElement (type, cx, cy, u, v, rot, atten);

  m_listPElem.push_front (pelem);

  // update phantom limits
  if (m_xmin > pelem->xmin())    m_xmin = pelem->xmin();
  if (m_xmax < pelem->xmax())    m_xmax = pelem->xmax();
  if (m_ymin > pelem->ymin())    m_ymin = pelem->ymin();
  if (m_ymax < pelem->ymax())    m_ymax = pelem->ymax();

  if (m_diameter < pelem->diameter())
    m_diameter = pelem->diameter();

  //  m_diameter = lineLength(m_xmin, m_ymin, m_xmax, m_ymax);

  m_nPElem++;
}


/*----------------------------------------------------------------------*/
/*			Input-Output Routines				*/
/*----------------------------------------------------------------------*/


/* NAME
 *   print				Print vertices of Phantom pelems
 *
 * SYNOPSIS
 *   print (phm)
 */

void 
Phantom::print (void) const
{
  printf("PRINTING Phantom\n\n");
  printf("number of pelems in Phantom = %d\n", m_nPElem);
  printf("limits: xmin=%8.2g  ymin=%8.2g  xmax=%8.2g  ymax=%8.2g\n",
	 m_xmin, m_ymin, m_xmax, m_ymax);
  
  for (PElemIterator i = m_listPElem.begin(); i != m_listPElem.end(); i++) {
      printf("PELEM:\n");
      printf("# pts=%3d	atten = %7.4f   rot = %7.2f (deg)\n",
	     (*i)->nOutlinePoints(), (*i)->atten(), convertRadiansToDegrees ((*i)->rot()));
    
    printf("xmin=%7.3g  ymin=%7.3g  xmax=%7.3g  ymax=%7.3g\n",
	   (*i)->xmin(), (*i)->ymin(), (*i)->xmax(), (*i)->ymax());
    
    //    for (int i = 0; i < m_nPoints; i++)
    //      printf("\t%8.3g    %8.3g\n", i->xOutline()[i], i->yOutline()[i]);
  }
}


/* NAME
 *   show		Show vector outline of Phantom to user
 *
 * SYNOPSIS
 *   show (pic)
 */

#ifdef HAVE_SGP
void 
Phantom::show (void) const
{
  double wsize = m_xmax - m_xmin;
  double xmin = m_xmin;
  double ymin = m_ymin;
  double xmax, ymax;
  SGP_ID gid;

  if ((m_ymax - m_ymin) > wsize)
      wsize = m_ymax - m_ymin;
  wsize *= 1.1;

  xmax = xmin + wsize;
  ymax = ymin + wsize; 
  
  printf("Drawing Phantom:\n\n");
  printf("    data limits: %9.3g, %9.3g, %9.3g, %9.3g\n",
	 m_xmin, m_ymin, m_xmax, m_ymax);
  printf("    window size: %9.3g, %9.3g, %9.3g, %9.3g\n",
	 xmin, ymin, xmax, ymax);

  gid = sgp2_init(0, 0, "Phantom Show");
  sgp2_window (xmin, ymin, xmax, ymax);

  draw();

  termgrf2();
}
#endif


/* NAME
 *   draw		Draw vector outline of Phantom
 *
 * SYNOPSIS
 *   draw ()
 */

#ifdef HAVE_SGP
void 
Phantom::draw (void) const
{
  for (PElemIterator i = m_listPElem.begin(); i != m_listPElem.end(); i++)
    sgp2_polyline_abs ((*i)->xOutline(), (*i)->yOutline(), (*i)->nOutlinePoints());
}
#endif


/* NAME
 *   std_rowland		Make head phantom of S.W. Rowland
 *
 * SYNOPSIS
 *   std_rowland ()
 *
 * REFERENCES
 *   S. W. Rowland, "Computer Implementation of Image Reconstruction
 *	Formulas", in "Image Reconstruction from Projections: Implementation
 *	and Applications", edited by G. T. Herman, 1978.
 */

void 
Phantom::std_rowland (void)
{
  addPElem("ellipse",  0.0000,  0.0000, 0.6900,  0.9200,   0.0,  1.00);
  addPElem("ellipse",  0.0000, -0.0184, 0.6624,  0.8740,   0.0, -0.98);
  addPElem("ellipse",  0.2200,  0.0000, 0.1100,  0.3100, -18.0, -0.02);
  addPElem("ellipse", -0.2200,  0.0000, 0.1600,  0.4100,  18.0, -0.02);
  addPElem("ellipse",  0.0000,  0.3500, 0.2100,  0.2500,   0.0,  0.01);
  addPElem("ellipse",  0.0000,  0.1000, 0.0460,  0.0460,   0.0,  0.01);
  addPElem("ellipse",  0.0000, -0.1000, 0.0460,  0.0460,   0.0,  0.01);
  addPElem("ellipse", -0.0800, -0.6050, 0.0460,  0.0230,   0.0,  0.01);
  addPElem("ellipse",  0.0000, -0.6050, 0.0230,  0.0230,   0.0,  0.01);
  addPElem("ellipse",  0.0600, -0.6050, 0.0230,  0.0230,   0.0,  0.01);
  addPElem("ellipse",  0.5538, -0.3858, 0.0330,  0.2060, -18.0,  0.03);
}

void 
Phantom::std_rowland_bordered (void)
{
  std_rowland ();
  addPElem ("ellipse", 0.000, 0.0000, 0.7500, 1.000, 0.0, 0.00);
}

/* NAME
 *   std_herman			Standard head phantom of G. T. Herman
 *
 * SYNOPSIS
 *   std_herman ()
 *
 * REFERENCES
 *   G. T. Herman, "Image Reconstructions from Projections:  The Fundementals
 *	of Computed Tomography", 1979.
 */

void 
Phantom::std_herman (void)
{
  addPElem("ellipse",  0.000,  1.50,  0.375, 0.3000,  90.00, -0.003);
  addPElem("ellipse",  0.675, -0.75,  0.225, 0.1500, 140.00,  0.010);
  addPElem("ellipse",  0.750,  1.50,  0.375, 0.2250,  50.00,  0.003);
  addPElem("segment",  1.375, -7.50,  1.100, 0.6250,  19.20, -0.204);
  addPElem("segment",  1.375, -7.50,  1.100, 4.3200,  19.21,  0.204);
  addPElem("segment",  0.000, -2.25,  1.125, 0.3750,   0.00, -0.003);
  addPElem("segment",  0.000, -2.25,  1.125, 3.0000,   0.00,  0.003);
  addPElem("segment", -1.000,  3.75,  1.000, 0.5000, 135.00, -0.003);
  addPElem("segment", -1.000,  3.75,  1.000, 3.0000, 135.00,  0.003);
  addPElem("segment",  1.000,  3.75,  1.000, 0.5000, 225.00, -0.003);
  addPElem("segment",  1.000,  3.75,  1.000, 3.0000, 225.00,  0.003);
  addPElem("triangle", 5.025,  3.75,  1.125, 0.5000, 110.75,  0.206);
  addPElem("triangle",-5.025,  3.75,  1.125, 0.9000,-110.75,  0.206);
  addPElem("ellipse",  0.000,  0.00,  8.625, 6.4687,  90.00,  0.416);
  addPElem("ellipse",  0.000,  0.00,  7.875, 5.7187,  90.00, -0.206);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS IDENTIFICATION
//
//      PhantomElement
//
// PURPOSE
//
////////////////////////////////////////////////////////////////////////////////////////////////////////


PhantomElement::PhantomElement (const char *type, const double cx, const double cy, const double u, const double v, const double rot, const double atten)
    : m_cx(cx), m_cy(cy), m_u(u), m_v(v), m_atten(atten), m_nPoints(0), m_xOutline(0), m_yOutline(0)
{
  m_rot = convertDegreesToRadians (rot);   // convert angle to radians

  if (strcasecmp (type, "rectangle") == 0)
    m_type = PELEM_RECTANGLE;
  else if (strcasecmp (type, "triangle") == 0)
    m_type = PELEM_TRIANGLE;
  else if (strcasecmp (type, "ellipse") == 0)
    m_type = PELEM_ELLIPSE;
  else if (strcasecmp (type, "sector") == 0)
    m_type = PELEM_SECTOR;
  else if (strcasecmp (type, "segment") == 0)
    m_type = PELEM_SEGMENT;
  else {
    sys_error (ERR_WARNING, "Unknown PhantomElement type %s [PhantomElement::PhantomElement]", type);
    m_type = PELEM_INVALID;
  }

  makeTransformMatrices ();     // calc transform matrices between phantom and normalized phantomelement
  makeVectorOutline ();		// calculate vector outline of pelem 

  // Find maximum diameter of Object
  double r2Max = 0;
  for (int i = 0; i < m_nPoints; i++) {
    double r2 = (m_xOutline[i] * m_xOutline[i]) + (m_yOutline[i] * m_yOutline[i]);
    if (r2 > r2Max)
      r2Max = r2;
  }
  m_diameter = 2 * sqrt( r2Max );

  m_rectLimits[0] = m_xmin;   m_rectLimits[1] = m_ymin;
  m_rectLimits[2] = m_xmax;   m_rectLimits[3] = m_ymax;
}


PhantomElement::~PhantomElement (void)
{
    delete m_xOutline;
    delete m_yOutline;
}

void 
PhantomElement::makeTransformMatrices (void)
{
  GRFMTX_2D temp;

  // To map normalized Pelem coords to world Phantom 
  //     scale by (u, v)				       
  //     rotate by rot					
  //     translate by (cx, cy)			       

  scale_mtx2 (m_xformObjToPhm, m_u, m_v);
  rot_mtx2  (temp, m_rot);
  mult_mtx2 (m_xformObjToPhm, temp, m_xformObjToPhm);
  xlat_mtx2 (temp, m_cx, m_cy);
  mult_mtx2 (m_xformObjToPhm, temp, m_xformObjToPhm);

  // to map world Phantom coodinates to normalized PElem coords
  //     translate by (-cx, -cy)
  //     rotate by -rot
  //     scale by (1/u, 1/v)

  xlat_mtx2 (m_xformPhmToObj, -m_cx, -m_cy);
  rot_mtx2  (temp, -m_rot);
  mult_mtx2 (m_xformPhmToObj, temp, m_xformPhmToObj);
  scale_mtx2 (temp, 1 / m_u, 1 / m_v);
  mult_mtx2 (m_xformPhmToObj, temp, m_xformPhmToObj);
}


/* NAME
 *   pelem_make_points		INTERNAL routine to calculate point array for an pelem
 *
 * SYNOPSIS
 *   makepelempts (pelem)
 *   PELEM *pelem	pelem whose points we are calculating
 *
 * NOTES
 *   Called by phm_add_pelem()
 */

static const double SCALE_PELEM_EXTENT=0.005;		// increase pelem limits by 0.5% 

void
PhantomElement::makeVectorOutline (void)
{
  double radius, theta, start, stop;
  double xfact, yfact;
  int cpts;

  m_nPoints = 0;
  switch (m_type) {
  case PELEM_RECTANGLE:
    m_nPoints = 5;
    m_xOutline = new double [m_nPoints];
    m_yOutline = new double [m_nPoints];
    m_xOutline[0] =-m_u;	m_yOutline[0] =-m_v;
    m_xOutline[1] = m_u;	m_yOutline[1] =-m_v;
    m_xOutline[2] = m_u;	m_yOutline[2] = m_v;
    m_xOutline[3] =-m_u; 	m_yOutline[3] = m_v;
    m_xOutline[4] =-m_u;	m_yOutline[4] =-m_v;
    break;
  case PELEM_TRIANGLE:
    m_nPoints = 4;
    m_xOutline = new double [m_nPoints];
    m_yOutline = new double [m_nPoints];
    m_xOutline[0] =-m_u;	m_yOutline[0] = 0.0;
    m_xOutline[1] = m_u;	m_yOutline[1] = 0.0;
    m_xOutline[2] = 0.0;	m_yOutline[2] = m_v;
    m_xOutline[3] =-m_u; 	m_yOutline[3] = 0.0;
    break;
  case PELEM_ELLIPSE:
    cpts = numCirclePoints (TWOPI);
    m_nPoints = cpts;
    m_xOutline = new double [m_nPoints];
    m_yOutline = new double [m_nPoints];
    calcEllipsePoints (m_xOutline, m_yOutline, cpts, m_u, m_v);
    break;
  case PELEM_SECTOR:
    radius = sqrt(m_u * m_u + m_v * m_v);
    theta = atan(m_u / m_v);		// angle with y-axis 
    start = 3.0 * HALFPI - theta;
    stop  = 3.0 * HALFPI + theta;
    cpts = numCirclePoints (stop - start);
    m_nPoints = 3 + cpts;
    m_xOutline = new double [m_nPoints];
    m_yOutline = new double [m_nPoints];
    
    m_xOutline[0] = 0.0;		m_yOutline[0] = m_v;
    m_xOutline[1] =-m_u;		m_yOutline[1] = 0.0;
    calcArcPoints (&m_xOutline[2], &m_yOutline[2], cpts, 0.0, m_v, radius, start, stop);
    m_xOutline[cpts + 2] = 0.0;
    m_yOutline[cpts + 2] = m_v;
    break;
  case PELEM_SEGMENT:
    radius = sqrt(m_u * m_u + m_v * m_v);
    theta = atan (m_u / m_v);		// angle with y-axis 
    start = 3.0 * HALFPI - theta;
    stop  = 3.0 * HALFPI + theta;
    
    cpts = numCirclePoints (stop - start);
    m_nPoints = cpts + 1;
    m_xOutline = new double [m_nPoints];
    m_yOutline = new double [m_nPoints];
    
    calcArcPoints (m_xOutline, m_yOutline, cpts, 0.0, m_v, radius, start, stop);
    m_xOutline[cpts] = -m_u;
    m_yOutline[cpts] = 0.0;
    break;
  default:
    sys_error(ERR_WARNING, "illegal pelem type %d [makeVectorOutline]", m_type);
    return;
  }
  
  rotate2d (m_xOutline, m_yOutline, m_nPoints, m_rot);
  xlat2d (m_xOutline, m_yOutline, m_nPoints, m_cx, m_cy);
  
  minmax_array (m_xOutline, m_nPoints, m_xmin, m_xmax);
  minmax_array (m_yOutline, m_nPoints, m_ymin, m_ymax);
  
  // increase pelem extent by SCALE_PELEM_EXTENT to eliminate chance of
  //   missing actual pelem maximum due to polygonal sampling 

  xfact = (m_xmax - m_xmin) * SCALE_PELEM_EXTENT;
  yfact = (m_ymax - m_ymin) * SCALE_PELEM_EXTENT;

  m_xmin -= xfact;
  m_ymin -= yfact;
  m_xmax += xfact;
  m_ymax += yfact;
}



/* NAME
 *   calc_arc			Calculate outline of a arc of a circle
 *
 * SYNOPSIS
 *   calc_arc (x, y, xcent, ycent, pts, r, start, stop)
 *   double x[], y[];		Array of points
 *   int pts			Number of points in array
 *   double xcent, ycent	Center of cirlce
 *   double r			Radius of circle
 *   double start, stop		Beginning & ending angles
 */

void 
PhantomElement::calcArcPoints (double x[], double y[], const int pts, const double xcent, const double ycent, const double r, const double start, const double stop)
{
    if (r <= 0.0)
	sys_error (ERR_WARNING, "negative or zero radius in calc_arc()");

    double theta = (stop - start) / (pts - 1);	// angle incr. between points 
    double c = cos(theta);
    double s = sin(theta);
  
    x[0] = r * cos (start) + xcent;
    y[0] = r * sin (start) + ycent;

    double xp = x[0] - xcent;
    double yp = y[0] - ycent;
    for (int i = 1; i < pts; i++) {
	double xc = c * xp - s * yp;
	double yc = s * xp + c * yp;
	x[i] = xc + xcent;
	y[i] = yc + ycent;
	xp = xc;  yp = yc;
    }
}


// NAME
//   PhantomElement::calcEllipsePoints    Calculate outline of a ellipse
//
// SYNOPSIS
//   calcEllipsePoints ()
//


void 
PhantomElement::calcEllipsePoints (double x[], double y[], const int pts, const double u, const double v)
{
    calcArcPoints (x, y, m_nPoints, 0.0, 0.0, 1.0, 0.0, TWOPI);   // make a unit circle 
    scale2d (x, y, m_nPoints, m_u, m_v);			     // scale to ellipse 
}


/* NAME
 *   circle_pts		Calculate number of points to use for circle segment
 *
 * SYNOPSIS
 *   n = circle_pts (theta)
 *   int n		Number of points to use for arc
 *   double theta	Length of arc in radians
 */

int 
PhantomElement::numCirclePoints (double theta) const
{
    if (theta < 0.0 || theta > TWOPI)
      sys_error(ERR_WARNING, "illegal values sent to circle_pts");

    return static_cast<int> (POINTS_PER_CIRCLE * theta / TWOPI + 1.5);
}


bool
PhantomElement::clipLineWorldCoords (double& x1, double& y1, double& x2, double &y2) const
{
  /* check if ray is outside of pelem extents */
  double cx1 = x1, cy1 = y1, cx2 = x2, cy2 = y2;
  if (! clip_rect (cx1, cy1, cx2, cy2, m_rectLimits))
    return false;
	
  // convert phantom coordinates to pelem coordinates 
  xform_mtx2 (m_xformPhmToObj, x1, y1);
  xform_mtx2 (m_xformPhmToObj, x2, y2);
	
  if (! clipLineNormalizedCoords (x1, y1, x2, y2))
    return false;

  // convert standard pelem coordinates back to phantom coordinates 
  xform_mtx2 (m_xformObjToPhm, x1, y1);
  xform_mtx2 (m_xformObjToPhm, x2, y2);

  return true;
}


/* NAME
 *   pelem_clip_line			Clip pelem against an arbitrary line
 *
 * SYNOPSIS
 *   pelem_clip_line (pelem, x1, y1, x2, y2)
 *   PhantomElement& pelem;		Pelem to be clipped
 *   double *x1, *y1, *x2, *y2	Endpoints of line to be clipped
 *
 * RETURNS
 *   true   if line passes through pelem
 *		(x1, y1, x2, y2 hold coordinates of new line)
 *   false  if line do not pass through pelem
 *		(x1, y1, x2, y2 are undefined)
 */

bool
PhantomElement::clipLineNormalizedCoords (double& x1, double& y1, double& x2, double& y2) const
{
  bool accept = false;

  switch (m_type) {
  case PELEM_RECTANGLE:
    double rect[4];
    rect[0] = -1.0;  rect[1] = -1.0;
    rect[2] = 1.0;  rect[3] = 1.0;
    accept = clip_rect (x1, y1, x2, y2, rect);
    break;
  case PELEM_ELLIPSE:
    accept = clip_circle (x1, y1, x2, y2, 0.0, 0.0, 1.0, 0.0, 0.0);
    break;
  case PELEM_TRIANGLE:
    accept = clip_triangle (x1, y1, x2, y2, 1.0, 1.0, true);
    break;
  case PELEM_SEGMENT:
    accept = clip_segment (x1, y1, x2, y2, m_u, m_v);
    break;
  case PELEM_SECTOR:
    accept = clip_sector (x1, y1, x2, y2, m_u, m_v);
    break;
  default:
    sys_error (ERR_WARNING, "Illegal pelem type %d [pelem_clip_line]", m_type);
    break;
  }

  return(accept);
}


// METHOD IDENTIFICATION 
//    PhantomElement::isPointInside		Check if point is inside pelem
//
// SYNOPSIS
//    is_point_inside (pelem, x, y, coord_type)
//    double x, y		Point to see if lies in pelem
//    int coord_type		Coordinate type (PELEM_COORD or PHM_COORD)
//
// RETURNS
//    true if point lies within pelem
//    false if point lies outside of pelem

bool
PhantomElement::isPointInside (double x, double y, const CoordType coord_type)
{
  if (coord_type == PHM_COORD) {
    xform_mtx2 (m_xformPhmToObj, x, y);
  } else if (coord_type != PELEM_COORD) {
    sys_error(ERR_WARNING, "Illegal coordinate type in pelem_is_point_inside");
    return (false);
  }

  switch (m_type) {
  case PELEM_RECTANGLE:
    if (x > 1. || x < -1. || y > 1. || y < -1.)
      return (false);
    else
      return (true);
    break;
  case PELEM_TRIANGLE:
    if (y < 0. || y > 1. - x || y > 1. + x)
      return (false);
    else
      return (true);
    break;
  case PELEM_ELLIPSE:
    if (x > 1. || x < -1. || y > 1. || y < -1.)
      return (false);
    if (x * x + y * y > 1.)		// check if inside unit circle
      return (false);
    else
      return (true);
    break;

    // for clipping segments & sectors, must NOT scale by (1/u, 1/v)
    // because this destroys information about size of arc component 

  case PELEM_SEGMENT:
    if (x > 1. || x < -1. || y > 0.)
	return (false);		// clip against y > 0 
    x *= m_u;			// put back u & v scale 
    y *= m_v;
    if (x * x + (y-m_v) * (y-m_v) > m_u * m_u + m_v * m_v)
      return (false);		// clip against circle, r = sqrt(@)
    else
      return (true);
    break;
  case PELEM_SECTOR:
      if (x > 1. || x < -1. || y > 1.)   // extent 
      return (false);
      if (y > 1. - x || y > 1. + x)      // triangle 	
	  return (false);		       // clip against triangle 
      x *= m_u;		       // circle: put back u & v scale 
    y *= m_v;
    if (x * x + (y-m_v) * (y-m_v) > m_u * m_u + m_v * m_v)
	return (false);		       // clip against circle 
    else
      return (true);
  break;
  default:
    sys_error (ERR_WARNING, "Illegal pelem type in pelem_is_point_inside()");
    break;
  }

  return (false);
}


