/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          if-2.cpp
**   Purpose:       Manipulate two image files
**   Programmer:    Kevin Rosenberg
**   Date Started:  May 2000
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: if-2.cpp,v 1.6 2000/06/19 17:58:13 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

/* FILE
 *   if-2.c         Generate a IF file from two input IF files
 */

#include "ct.h"

enum {O_ADD, O_SUB, O_MUL, O_COMP, O_VERBOSE, O_HELP, O_VERSION};

static struct option my_options[] =
{
  {"add", 0, 0, O_ADD},
  {"sub", 0, 0, O_SUB},
  {"mul", 0, 0, O_MUL},
  {"comp", 0, 0, O_COMP},
  {"verbose", 0, 0, O_VERBOSE},
  {"help", 0, 0, O_HELP},
  {"version", 0, 0, O_VERSION},
  {0, 0, 0, 0}
};

void 
if2_usage (const char *program)
{
  cout << "usage: " << fileBasename(program) << " infile1 infile2 outfile [OPTIONS]" << endl;
  cout << "Generate an image file from two input image files files" << endl;
  cout << endl;
  cout << "     infile1    Name of first input IF file" << endl;
  cout << "     infile2    Name of second input IF file" << endl;
  cout << "     outfile    Name of output IF file" << endl;
  cout << "     --add      Add images" << endl;
  cout << "     --sub      Subtract image 2 from image 1" << endl;
  cout << "     --mul      Multiply images" << endl;
  cout << "     --comp     Compare images" << endl;
  cout << "     --verbose  Verbose modem" << endl;
  cout << "     --version  Print version" << endl;
  cout << "     --help     Print this help message" << endl;
}

int 
if2_main (int argc, char *const argv[])
{
  ImageFile* pim_in1;
  ImageFile* pim_in2;
  ImageFile* pim_out;
  char *in_file1;
  char *in_file2;
  char *out_file;
  int opt_verbose = 0;
  int opt_add = 0;
  int opt_sub = 0;
  int opt_mul = 0;
  int opt_comp = 0;

  while (1) {
    int c = getopt_long (argc, argv, "", my_options, NULL);
      
    if (c == -1)
      break;
      
    switch (c) {
    case O_ADD:
      opt_add = 1;
      break;
    case O_SUB :
      opt_sub = 1;
      break;
    case O_MUL:
      opt_mul = 1;
      break;
    case O_COMP:
      opt_comp = 1;
      break;
    case O_VERBOSE:
      opt_verbose = 1;
      break;
    case O_VERSION:
#ifdef VERSION
      cout << "Version " << VERSION << endl;
#else
      cout << "Unknown version number" << endl;
#endif
      return (0);
    case O_HELP:
    case '?':
      if2_usage(argv[0]);
      return (0);
    default:
      if2_usage(argv[0]);
      return (1);
    }
  }
  
  if (optind + 3 != argc) {
    if2_usage(argv[0]);
    return (1);
  }
  
  in_file1 = argv[optind];
  in_file2 = argv[optind + 1];
  out_file = argv[optind + 2];

  pim_in1 = new ImageFile (in_file1);
  pim_in2 = new ImageFile (in_file2);
  ImageFile& im_in1 = *pim_in1;
  ImageFile& im_in2 = *pim_in2;

  if (! im_in1.fileRead() || ! im_in2.fileRead()) {
      sys_error (ERR_WARNING, "Error reading an image");
      return (1);
  }

  if (im_in1.nx() != im_in2.nx() || im_in1.ny() != im_in2.ny()) {
    sys_error (ERR_SEVERE, "Error: Size of image 1 (%d,%d) and image 2 (%d,%d) do not match",
	    im_in1.nx(), im_in1.ny(), im_in2.nx(), im_in2.ny());
    return(1);
  }
  if (im_in1.nx() < 0 || im_in1.ny() < 0) {
      sys_error (ERR_SEVERE, "Error: Size of image < 0");
      return(1);
  }

  pim_out = new ImageFile (out_file, im_in1.nx(), im_in1.ny());
  ImageFile& im_out = *pim_out;
  if (! im_out.fileCreate()) {
    sys_error (ERR_WARNING, "Could not open output file %s", out_file);
    return (1);
  }

  string strOperation;
  ImageFileArray v1 = im_in1.getArray();
  ImageFileArray v2 = im_in2.getArray();
  ImageFileArray vout = im_out.getArray();

  if (opt_add) {
    strOperation = "Add Images";
    for (int ix = 0; ix < im_in1.nx(); ix++) {
      ImageFileColumn in1 = v1[ix];
      ImageFileColumn in2 = v2[ix];
      ImageFileColumn out = vout[ix];
      for (int iy = 0; iy < im_in1.ny(); iy++)
	*out++ = *in1++ + *in2++;
    }
  } else if (opt_sub) {
    strOperation = "Subtract Images";
    for (int ix = 0; ix < im_in1.nx(); ix++) {
      ImageFileColumn in1 = v1[ix];
      ImageFileColumn in2 = v2[ix];
      ImageFileColumn out = vout[ix];
      for (int iy = 0; iy < im_in1.ny(); iy++)
	  *out++ = *in1++ - *in2++;
    }
  } else if (opt_mul) {
    strOperation = "Multiply Images";
    for (int ix = 0; ix < im_in1.nx(); ix++) {
      ImageFileColumn in1 = v1[ix];
      ImageFileColumn in2 = v2[ix];
      ImageFileColumn out = vout[ix];
      for (int iy = 0; iy < im_in1.ny(); iy++)
	*out++ = *in1++ * *in2++;
    }
  } else if (opt_comp) {
    double abs_error = 0.;
    strOperation = "Subtract Images";
    for (int ix = 0; ix < im_in1.nx(); ix++) {
      ImageFileColumn in1 = v1[ix];
      ImageFileColumn in2 = v2[ix];
      ImageFileColumn out = vout[ix];
      for (int iy = 0; iy < im_in1.ny(); iy++) {
	double diff = *in1++ - *in2++;
	*out++ = diff;
	abs_error += fabs(diff);
      }
    }
    abs_error /= (im_in1.nx() * im_in1.ny());
    cout << "Average Error: " << abs_error << endl;
  }

  im_out.arrayDataWrite();
  im_out.labelsCopy (im_in1, "if-2 file 1: ");
  im_out.labelsCopy (im_in2, "if-2 file 2: ");
  im_out.labelAdd (Array2dFileLabel::L_HISTORY, strOperation.c_str());

  im_out.fileClose();

  return (0);
}

#ifndef NO_MAIN
int 
main (int argc, char *const argv[])
{
  int retval = 1;

  try {
    retval = if2_main(argc, argv);
  } catch (exception e) {
    cerr << "Exception: " << e.what() << endl;
  } catch (...) {
    cerr << "Unknown exception" << endl;
  }

  return (retval);
}
#endif

